/*******************/
/* System includes */
/*******************/

#include <stdlib.h>
#include <stddef.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <ctype.h>
#include <limits.h>
#include <unistd.h>

/******************/
/* EPICS includes */
/******************/

/* EPICS includes */
#include <drvIpac.h>
#include <iocsh.h>
#include <epicsExport.h>
#include <epicsThread.h>
#include <epicsString.h>
#include <epicsTimer.h>
#include <epicsExit.h>
#include <epicsMutex.h>
#include <epicsMessageQueue.h>
#include <cantProceed.h>
#include <devLib.h>
#include <errlog.h>
#include <asynPortDriver.h>

//for file writing
#include <sys/types.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <macLib.h>
#include	<time.h>
#include	<dbDefs.h>
#include	<cadef.h>		/* includes dbAddr.h */
#include	<epicsPrint.h>
#include	<callback.h>
#include	<epicsEvent.h>
#include	<epicsTime.h>

#ifndef DRVHYFWUpgradeIPASYN_H
#define DRVHYFWUpgradeIPASYN_H
#endif

/**************************************************************/
/* A couple of key parameters defining the ADC memory layout. */
#define BANKSIZE  				160000        	/* more than 45 + 567 * 132 word */
#define MAX_ADC_CHANNELS    	1	      		/* number of ADC channels on FWUpgrade card */

/* Identifiers in PROM. */
#define VITA4       			"VITA 4 "
#define HYTECID_LO  			0x0300
#define HYTECID_HI  			0x0080
#define PROM_MODEL  			0xFWUpgrade

/* define bits in the FWUpgrade CSR */
#define CSR_HF      0x0001          /* Set when lower conversion memory filled. */
#define CSR_F       0x0002          /* Set when upper conversion memory filled. */     
#define CSR_CC      0x0004          /* Set when requested sample count captured. */
#define CSR_MIS     0x0008          /* Set when memory update is inhibited. */
#define CSR_MII     0x0010          /* Set if interrupt generated by inhibit (MIS). */
#define CSR_EII     0x0020          /* Set to enable MIS interrupt. */
#define CSR_DA      0x0040          /* Set to allow the unit to disarm on completion of memory acquisition */
#define CSR_1M      0x0080          /* 1 => 128K samples, 0 => 256K samples per channel. */
#define CSR_HE      0x0100          /* Enable interrupt on Half Full set. */
#define CSR_FE      0x0200          /* Enable interrupt on Full set. */
#define CSR_EE      0x0400          /* Enable end of sample sequence interrupt. */
#define CSR_ET      0x0800          /* Allow external enable to control ADC. */
#define CSR_XC      0x1000          /* 0 => use internal clock, 1 => external clock. */
#define CSR_ST      0x2000          /* Set to trigger capture of programmed samples. */
#define CSR_EX      0x4000          /* 0 => continuous capture, 1 => on trigger. */
#define CSR_ARM     0x8000          /* Set to enable capture into waveform RAM. */

#define EXTCSR_TEN  0x0001          /* Triggered status */
#define EXTCSR_SMI  0x0002          /* Software inhibit control and status. Write 1 to inhibit memory writing; write 0 to allow. Read 1 means hardware inhibit; read 0 means hardware allows */
#define EXTCSR_IS   0x0004          /* Single conversion completed flag */
#define EXTCSR_ISE  0x0008          /* Single conversion completed interrupt enable */
#define EXTCSR_CONT 0x0010          /* Allow multiple triggers on CC mode */
#define EXTCSR_R  	0x0040          /* ADC range 0 = +/-10V and 1 = +/-5V */
#define EXTCSR_TE  	0x1000          /* Set 0 = external trigger on rising edge, 1 = external trigger on falling edge */
#define EXTCSR_BTE 	0x2000          /* Set 1 = extrenal trigger on both edge */
#define EXTCSR_CAL	0x4000          /* Set 1 = do not use firmware calibration, 0 = use firmware calibration. No meaning when EX = 1 */

#define TRCSR_TR  	0x8000          /* Transient recorder enable */
#define TRCSR_VT  	0x4000          /* Transient recorder voltage trigger mode */
#define TRCSR_SOL  	0x0020          /* Transient recorder trigger output edge 0=high, 1=low */
#define TRCSR_SIL  	0x0010          /* Transient recorder stop in edge. 0=high, 1=low */
#define TRCSR_ST  	0x0008          /* Transient recorder stop status */
#define TRCSR_SS  	0x0004          /* Transient recorder software stop bit */
#define TRCSR_CHL  	0x0003          /* Choice a channel for VT mode */

#ifndef USHORT 
#define USHORT unsigned short
#endif
#ifndef ULONG
#define ULONG unsigned int
#endif

/*****************************************************************************/
/*                                                                           */
/*                              Data Structures                              */
/*                                                                           */
/*****************************************************************************/


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/* ADC Register and P#define hyFWUpgradeSupportString			"SUPPORT"
ROM Layout                                              */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/* The I/O memory is divided into two areas: register and PROM. We separate
 * them out for coding convenience. 
 */

typedef struct
{
    USHORT status;                  /* 0x00 */
    USHORT pointer[2];              /* 0x02 - 0x04 */
    USHORT clockRate;               /* 0x06 */
    USHORT vector;                  /* 0x08 */
    USHORT extStatus;               /* 0x0A */
    USHORT unused[10];              /* 0x0C~ */
} REGISTERS;


/* This data structure maps the image of the ID PROM.  The PROM
 * starts at offset 0x80 from the base of the register memory. */
typedef struct
{
    char vita4[6];                  /* 0x80 */
    USHORT hytecIdHigh;             /* 0x86 */
    USHORT hytecIdLow;              /* 0x88 */
    USHORT model;                   /* 0x8A */
    USHORT revision;                /* 0x8C */
    USHORT unused1;                 /* 0x8E */
    USHORT driverIdHigh;            /* 0x90 */
    USHORT driverIdLow;             /* 0x92 */
    USHORT flags;                   /* 0x94 */
    USHORT byteCount;               /* 0x96 */

    USHORT calType;                 /* 0x98 */
    USHORT serialNumber;            /* 0x9A */
    USHORT unused2;                 /* 0x9C */
    USHORT WLO;                     /* 0x9E */
} ID_PROM;

/* Now put them together, the I/O memory map. */
typedef struct
{
    REGISTERS reg;                              /* 0x00 - 0x2F */
    char unused[0x80-sizeof(REGISTERS)];        /* 0x30 - 0x80 */
    ID_PROM prom;                               /* 0x80 - 0x9F */
} IO_MEMORY;


/*****************************************************************************/
/*                                                                           */
/*                      Interrupt/Queue Routines	                         */
/*                                                                           */
/*****************************************************************************/

/** This is the class definition for the HyFWUpgradeAsyn class
  */
class HyFWUpgradeAsyn : public asynPortDriver {
public:
	HyFWUpgradeAsyn(char *portName, int carrierNum, int ipSlotNum, char *fileName, int mode);
    virtual asynStatus readInt32(asynUser *pasynUser, epicsInt32 *value);

protected:
    int hyFWUpgradeError_;				//ERROR
    #define FIRST_FWUpgrade_PARAM hyFWUpgradeError_
	int hyFWUpgradeDriverVersion_;
    #define LAST_FWUpgrade_PARAM hyFWUpgradeDriverVersion_

private:
	    
	volatile IO_MEMORY 	*ipio;			   	/* pointer to IP card A16 space registers, ID and calibration data */ 
	volatile USHORT 	*ipmem;   	/* pointer to IP card memory space */

	int 	openFile(char * fileName);
	int 	checkFilename(char * fileName, int model);
	int 	writeFlash();
	int 	readBack();
	int 	compare(int fullText, int pause);
	int		clearBuffer();

	int 	memTest(int size);

	unsigned char 	buf[BANKSIZE];					/* store byte binary array */
	USHORT 	wordbuf[BANKSIZE/2];			/* store word binary array */
	USHORT 	verifyBuf[BANKSIZE/2];			/* buffer for read back */

	/* Storage for the current Configuration parameters */
    USHORT	carrierNum;		                /* carrier index number when adding carrier */
    USHORT	ipSlotNum;		                /* IP card slot number on IP carrier - boot time configuration only */
    int		mode;		                	/* mode */

    int fileLength;
	int error;
};

#define hyFWUpgradeErrorString	    	"ERROR"				 
#define hyFWUpgradeDriverVersionString	"DRIVERVERSION"

#define NUM_PARAMS (&LAST_FWUpgrade_PARAM - &FIRST_FWUpgrade_PARAM + 1)

/** Configuration command, called directly or from iocsh */
/* 
 * Where: -
 *     portName = asyn port name
 *   carrierNum = carrier index number when adding a carrier card by ipacAddHy8002 call
 *    ipSlotNum = IP slot number on the carrier board (0 = A, 1 = B, 2 = C, 3 = D, 4 = E, 5 = F),
 *     fileName = firmware binary file name. Ignored in mode 0 and 1.
 *     	   mode = 0:memory test
 * 				  1:View firmware 
 *			  	  2:Read PROM and compare with specified .bit file, 
 * 				  3:Firmware upgrade
 */
extern "C" int HyFWUpgradeAsynInit(char *portName, int vmeSlotNum, int ipSlotNum, char *fileName, int mode)
{
    HyFWUpgradeAsyn *pHyFWUpgradeAsyn = new HyFWUpgradeAsyn(portName, vmeSlotNum, ipSlotNum, fileName, mode);
    pHyFWUpgradeAsyn = NULL;  /* This is just to avoid compiler warnings */
    return(asynSuccess);
}

			
// Init
static const iocshArg initArg0 = { "Port name",iocshArgString};
static const iocshArg initArg1 = { "vmeSlotNum",iocshArgInt};
static const iocshArg initArg2 = { "ipSlotNum",iocshArgInt};
static const iocshArg initArg3 = { "fileName",iocshArgString};
static const iocshArg initArg4 = { "mode",iocshArgInt};
static const iocshArg * const initArgs[] = {&initArg0,
                                             &initArg1,
                                             &initArg2,
                                             &initArg3,
                                             &initArg4};
static const iocshFuncDef initFuncDef = {"HyFWUpgradeAsynInit",5,initArgs};
static void initCallFunc(const iocshArgBuf *args)
{
    HyFWUpgradeAsynInit(args[0].sval, args[1].ival, args[2].ival,
					args[3].sval, args[4].ival);
}


void drvHyFWUpgradeAsynRegister(void)
{
    iocshRegister(&initFuncDef,initCallFunc);
}

extern "C" {
epicsExportRegistrar(drvHyFWUpgradeAsynRegister);
}

